# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import warnings
from itertools import chain

import numpy as np
from math import sqrt
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm, tqdm_notebook

from qucumber.callbacks import CallbackList
from qucumber.binary_rbm import BinaryRBM
import qucumber.utils.cplx as cplx
import qucumber.utils.unitaries as unitaries
from qucumber.callbacks import CallbackList

__all__ = [
    "ComplexWavefunction"
]

class ComplexWavefunction(object):
    
    def __init__(self, unitary_dict, num_visible,
                 num_hidden, gpu=True):
        super(ComplexWavefunction, self).__init__()
        self.num_visible = int(num_visible)
        self.num_hidden = int(num_hidden)
        self.rbm_am = BinaryRBM(num_visible, num_hidden, gpu=gpu)
        self.rbm_ph = BinaryRBM(num_visible, num_hidden, gpu=gpu)
              
        self.size_cut=20
        self.space = None
        self.Z = 0.0
 
        self.networks = ["rbm_am","rbm_ph"]
        self.device = self.rbm_am.device
        self.unitary_dict = unitary_dict 
        self.visible_state = torch.zeros(1,self.num_visible,
                                         device=self.rbm_am.device,
                                         dtype=torch.double)
        self.hidden_state = torch.zeros(1,self.num_hidden,
                                         device=self.rbm_am.device,
                                         dtype=torch.double)

    def initialize_parameters(self):
        r"""Randomize the parameters :math:`\bm{\lambda}=\{\bm{W}^{\bm{\lambda}},
        \bm{b}^{\bm{\lambda}},\bm{c}^{\bm{\lambda}}\}` and :math:`\bm{\mu}=\{\bm{W}^{\bm{\mu}},
        \bm{b}^{\bm{\mu}},\bm{c}^{\bm{\mu}}\}` of the amplitude and phase RBM respectively."""
        self.rbm_am.initialize_parameters()
        self.rbm_ph.initialize_parameters()
        
    def set_visible_layer(self,v):
        r""" Set the visible state to a given vector/matrix
        
        :param v: State to initialize the wavefunction to
        :type v: torch.Tensor
        """
        self.visible_state = v
   
    def amplitude(self,v):
        r""" Compute the amplitude of a given vector/matrix of visible states:
        
        .. math::
            
            \text{amplitude}(\bm{\sigma})=|\psi_{\bm{\lambda\mu}}(\bm{\sigma})|=
            e^{-\mathcal{E}_{\bm{\lambda}}(\bm{\sigma})/2}

        :param v: visible states :math:`\bm{\sigma}`
        :type v: torch.tensor

        :returns Matrix/vector containing the amplitudes of v
        :rtype torch.tensor
        """
        return (-self.rbm_am.effective_energy(v)).exp().sqrt()
    
    def phase(self,v):
        r""" Compute the phase of a given vector/matrix of visible states

        .. math::
            
            \text{phase}(\bm{\sigma})=-\mathcal{E}_{\bm{\mu}}(\bm{\sigma})/2

        :param v: visible states :math:`\bm{\sigma}`
        :type v: torch.tensor

        :returns Matrix/vector containing the phases of v
        :rtype torch.tensor
        """
        return -0.5*self.rbm_ph.effective_energy(v)
   
    def psi(self,v):
        r""" Compute the wavefunction of a given vector/matrix of visible states:

        .. math::

            \psi_{\bm{\lambda\mu}}(\bm{\sigma}) = e^{-[\mathcal{E}_{\bm{\lambda}}(\bm{\sigma})+i\mathcal{E}_{\bm{\mu}}(\bm{\sigma}]/2}

        :param v: visible states :math:`\bm{\sigma}`
        :type v: torch.tensor

        :returns: Complex object containing the value of the wavefunction for each visible state
        :rtype: torch.tensor
        """
        cos_phase = (self.phase(v)).cos() 
        sin_phase = (self.phase(v)).sin() 
        psi = torch.zeros(2, dtype=torch.double, device=self.device)
        psi[0] = self.amplitude(v)*cos_phase 
        psi[1] = self.amplitude(v)*sin_phase
        return psi

    def gradient(self,basis,sample):
        r"""Compute the gradient of a set (v_state) of samples, measured
        in different bases
        
        :param basis: A set of basis, (i.e.vector of strings)
        :type basis: np.array
        """
        num_U = 0               # Number of 1-local unitary rotations
        rotated_sites = []      # List of site where the rotations are applied
        grad = []               # Gradient
        
        # Read where the unitary rotations are applied
        for j in range(self.num_visible):
            if (basis[j] != 'Z'):
                num_U += 1
                rotated_sites.append(j)

        # If the basis is the reference one ('ZZZ..Z')
        if (num_U == 0):
            grad.append(self.rbm_am.effective_energy_gradient(sample))  # Real 
            grad.append(0.0)                                            # Imaginary
        
        else:
            # Initialize
            vp = torch.zeros(self.num_visible, dtype=torch.double, device = self.device)
            rotated_grad = [torch.zeros(2,self.rbm_am.num_pars,dtype=torch.double, device = self.device),torch.zeros(2,self.rbm_ph.num_pars,dtype=torch.double, device = self.device)]
            Upsi = torch.zeros(2, dtype=torch.double, device = self.device)
            
            # Sum over the full subspace where the rotation are applied
            #sub_state = self.generate_visible_space(num_U)
            sub_space = self.generate_Hilbert_space(num_U)
            for x in range(1<<num_U):
                # Create the correct state for the full system (given the data)
                cnt = 0
                for j in range(self.num_visible):
                    if (basis[j] != 'Z'):
                        #vp[j]=sub_state[x][cnt] # This site sums (it is rotated)
                        vp[j] = sub_space[x][cnt]
                        cnt += 1
                    else:
                        vp[j]=sample[j]         # This site is left unchanged

                U = torch.tensor([1., 0.], dtype=torch.double, device = self.device) #Product of the matrix elements of the unitaries
                for ii in range(num_U):
                    tmp = self.unitary_dict[basis[rotated_sites[ii]]][:,int(sample[rotated_sites[ii]]),int(vp[rotated_sites[ii]])]
                    U = cplx.scalar_mult(U,tmp.to(self.device))
                
                # Gradient on the current configuration
                grad_vp = [self.rbm_am.effective_energy_gradient(vp),self.rbm_ph.effective_energy_gradient(vp)]
                
                # NN state rotated in this bases
                Upsi_v = cplx.scalar_mult(U,self.psi(vp))
                
                Upsi += Upsi_v            
                rotated_grad[0] += cplx.scalar_mult(Upsi_v,cplx.make_complex(grad_vp[0],torch.zeros_like(grad_vp[0])))
                rotated_grad[1] += cplx.scalar_mult(Upsi_v,cplx.make_complex(grad_vp[1],torch.zeros_like(grad_vp[1]))) 

            grad.append(cplx.scalar_divide(rotated_grad[0],Upsi)[0,:])
            grad.append(-cplx.scalar_divide(rotated_grad[1],Upsi)[1,:])

        return grad

    def sample(self, k_cd):
        r"""Performs k steps of Block Gibbs sampling. One step consists of sampling
        the hidden state :math:`\bm{h}` from the conditional distribution 
        :math:`p_{\bm{\lambda}}(\bm{h}\:|\:\bm{v})`, and sampling the visible state
        :math:`\bm{v}` from the conditional distribution :math:`p_{\bm{\lambda}}(\bm{v}\:|\:\bm{h})`.

        :param k_cd: Number of Block Gibbs steps.
        :type k_cd: int
        """
        for _ in range(k_cd):
            self.hidden_state = self.rbm_am.sample_h_given_v(self.visible_state)
            self.visible_state = self.rbm_am.sample_v_given_h(self.hidden_state)

    def generate_Hilbert_space(self,size):
        r"""Generates Hilbert space of dimension :math:`2^{\text{size}}`.
    
        :returns: A tensor with all the basis states of the Hilbert space.
        :rtype: torch.Tensor
        """
        if (size > self.size_cut):
            raise ValueError('Size of the Hilbert space too large!')
        else: 
            space = torch.zeros((1 << size, size),
                                device=self.device, dtype=torch.double)
            for i in range(1 << size):
                d = i
                for j in range(size):
                    d, r = divmod(d, 2)
                    space[i, size - j - 1] = int(r)
            return space

    def compute_normalization(self):
        r"""Compute the normalization constant of the wavefunction.
        
        .. math::
            
            Z_{\bm{\lambda}}=\sqrt{\sum_{\bm{\sigma}}|\psi_{\bm{\lambda\mu}}|^2}=
            \sqrt{\sum_{\bm{\sigma}} p_{\bm{\lambda}}(\bm{\sigma})}

        :param space: A rank 2 tensor of the entire visible space.
        :type space: torch.Tensor

        """
        if (self.space is None):
            raise ValueError('Missing Hilbert space')
        else:
            self.Z = self.rbm_am.compute_partition_function(self.space)

    def save(self, location, metadata={}):
        """Saves the RBM parameters to the given location along with
        any given metadata.

        :param location: The location to save the RBM parameters + metadata
        :type location: str or file
        :param metadata: Any extra metadata to store alongside the RBM
                         parameters
        :type metadata: dict
        """
        # add extra metadata to dictionary before saving it to disk
        data = {"rbm_am":self.rbm_am.state_dict(),"rbm_ph":self.rbm_ph.state_dict(), **metadata}
        torch.save(data, location)

    def load(self, location):
        """Loads the RBM parameters from the given location ignoring any
        metadata stored in the file. Overwrites the RBM's parameters.

        .. note::
            The RBM object on which this function is called must
            have the same shape as the one who's parameters are being
            loaded.

        :param location: The location to load the RBM parameters from
        :type location: str or file
        """
        try:
            state_dict = torch.load(location)
        except AssertionError as e:
            state_dict = torch.load(location, lambda storage, loc: 'cpu')

        self.rbm_am.load_state_dict(state_dict['rbm_am'], strict=False)
        self.rbm_ph.load_state_dict(state_dict['rbm_ph'], strict=False)



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os.path
from pathlib import Path

import torch

from .callback import Callback


class ModelSaver(Callback):
    """Callback which allows model parameters (along with some metadata)
    to be saved to disk at regular intervals.

    This Callback is called at the end of each epoch. If `save_initial` is
    `True`, will also be called at the start of the training cycle.

    :param period: Frequency of model saving (in epochs).
    :type period: int
    :param folder_path: The directory in which to save the files
    :type folder_path: str
    :param file_name: The name of the output files. Should be a format string
                      with one blank, which will be filled with either the
                      epoch number or the word "initial".
    :type file_name: str
    :param save_initial: Whether to save the initial parameters (and metadata).
    :type save_initial: bool
    :param metadata: The metadata to save to disk with the model parameters
                     Can be either a function or a dictionary. In the case of a
                     function, it must take 2 arguments the RBM being trained,
                     and the current epoch number, and then return a dictionary
                     containing the metadata to be saved.
    :type metadata: callable or dict or None
    :param metadata_only: Whether to save *only* the metadata to disk.
    :type metadata_only: bool
    """

    def __init__(self, period, folder_path, file_name,
                 save_initial=True,
                 metadata=None, metadata_only=False):
        self.folder_path = folder_path
        self.period = period
        self.file_name = file_name
        self.save_initial = save_initial
        self.metadata = metadata
        self.metadata_only = metadata_only

        self.path = Path(folder_path)
        self.path.mkdir(parents=True, exist_ok=True)
        self.path = self.path.resolve()

    def _save(self, rbm, epoch, save_path):
        if callable(self.metadata):
            metadata = self.metadata(rbm, epoch)
        elif isinstance(self.metadata, dict):
            metadata = self.metadata
        elif self.metadata is None:
            metadata = {}

        if self.metadata_only:
            torch.save(metadata, save_path)
        else:
            rbm.save(save_path, metadata)

    def on_train_start(self, rbm):
        if self.save_initial:
            save_path = os.path.join(self.path,
                                     self.file_name.format("initial"))
            self._save(rbm, 0, save_path)

    def on_epoch_end(self, rbm, epoch):
        if epoch % self.period == 0:
            save_path = os.path.join(self.path, self.file_name.format(epoch))
            self._save(rbm, epoch, save_path)



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import torch
import numpy as np

def load_data(tr_samples_path,tr_psi_path=None,tr_bases_path=None,bases_path=None):
    r"""A function that will load in the data required for training.
    
    :param tr_samples_path: The path to the training data.
    :type tr_samples_path: str
    :param tr_psi_path: The path to the target/true wavefunctio (default = None).
    :type tr_psi_path: str
    :param tr_bases_path: The path to the basis data.
    :type tr_bases_path: str
    :param bases_path: The path to a file containing all possible bases used in 
                       the tr_bases_path file.
    :type bases_path: str

    :returns: A list of all input parameters.
    :rtype: list
    """
    data = [] 
    data.append(torch.tensor(np.loadtxt(tr_samples_path, dtype= 'float32'),dtype=torch.double))

    if tr_psi_path is not None:
        target_psi_data = np.loadtxt(tr_psi_path, dtype= 'float32')
        target_psi = torch.zeros(2,len(target_psi_data), dtype=torch.double)
        target_psi[0] = torch.tensor(target_psi_data[:,0], dtype=torch.double)
        target_psi[1] = torch.tensor(target_psi_data[:,1], dtype=torch.double)
        data.append(target_psi)

    if tr_bases_path is not None:
        data.append(np.loadtxt(tr_bases_path,dtype=str))

    if bases_path is not None:
        bases_data = np.loadtxt(bases_path,dtype=str)
        bases = []
        for i in range(len(bases_data)):
            tmp = ""
            for j in range(len(bases_data[i])):
                if bases_data[i][j] is not " ":
                    tmp += bases_data[i][j]
            bases.append(tmp)
        data.append(bases)
    return data


def extract_refbasis_samples(train_samples,train_bases):
    r"""A function that extracts the reference basis samples in the data.

    :param train_samples: The training samples.
    :type train_samples: numpy.array
    :param train_bases: The bases of the training samples.
    :type train_bases: numpy.array

    :returns: The samples in the data that are only in the reference basis.
    :rtype: torch.tensor
    """
    tmp = []
    num_visible=train_samples.shape[-1]
    for i in range(train_samples.shape[0]):
        flag = 0
        for j in range(num_visible):
            if train_bases[i][j] != 'Z':
                flag = 1
                break
        if flag == 0:
            tmp.append(train_samples[i])
    z_samples = torch.zeros(len(tmp),num_visible,dtype=torch.double)
    for i in range(len(tmp)):
        for j in range(num_visible):
            z_samples[i][j] = tmp[i][j]
    return z_samples



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import torch

"""
A module that allows torch to handle complex algebra.
---------
SYNTAX / ORDERING OF INDICES
matrices / tensors: m[2][i][j] >>> 2 = real and imaginary part
                               >>> i = number of rows in the real and
                                       imaginary parts
                               >>> j = number of columns in the real and
                                       imaginary parts

vectors: v[2][i]               >>> 2 = real and imaginary part
                               >>> i = number of rows in the real and
                                       imaginary parts

scalars: s[2]                  >>> 2 = real and imaginary part
---------
"""


def make_complex(x, y):  # wont work for scalars
    """A function that combines the real (x) and imaginary (y) parts of a
       vector or a matrix.

    ..note:: x and y must have the same shape. Also, this will not work for
             rank zero tensors (scalars).

    :param x: A vector or matrix.
    :type x: torch.tensor

    :param y: A vector or matrix.
    :type y: torch.tensor

    :returns: The tensor [x,y].
    :rtype: torch.tensor
    """
    if (len(list(x.size())) == 2) and (len(list(y.size())) == 2):  # matrices
        z = torch.zeros(2, x.size()[0], x.size()[1],
                        dtype=torch.double, device=x.device)
        z[0] = x
        z[1] = y
    else:  # vectors
        z = torch.zeros(2, x.size()[0], dtype=torch.double, device=x.device)
        z[0] = x
        z[1] = y
    return z


# replaces MS mult, VS mult. x has to be scalar, y has to be matrix
def scalar_mult(x, y):
    """A function that computes the product between complex matrices and scalars,
       complex vectors and scalars or two complex scalars.

    ..note:: If one wished to do vector-scalar multiplication or matrix-scalar
             multiplication, you must put the vector / matrix as the first
             argument (x).

    :param x: A complex scalar, vector or matrix.
    :type x: torch.tensor

    :param y: A complex scalar, vector or matrix.
    :type y: torch.tensor

    :returns: The product between x and y.
    :rtype: torch.tensor
    """
    z = torch.zeros_like(y)
    z[0] = x[0]*y[0] - x[1]*y[1]
    z[1] = x[0]*y[1] + x[1]*y[0]

    return z


def matmul(x, y):
    """A function that computes complex matrix-matrix and matrix-vector products.

    ..note:: If one wishes to do matrix-vector products, the vector must be the
             second argument (y).

    :param x: A complex matrix.
    :type x: torch.tensor

    :param y: A complex vector or matrix.
    :type y: torch.tensor

    :returns: The product between x and y.
    :rtype: torch.tensor
    """
    if len(list(y.size())) == 2:
        # if one of them is a vector (i.e. wanting to do MV mult)
        z = torch.zeros(2, x.size()[1], dtype=torch.double, device=x.device)
        z[0] = torch.mv(x[0], y[0]) - torch.mv(x[1], y[1])
        z[1] = torch.mv(x[0], y[1]) + torch.mv(x[1], y[0])

    if len(list(y.size())) == 3:
        z = torch.zeros(2, x.size()[1], y.size()[2],
                        dtype=torch.double, device=x.device)
        z[0] = torch.matmul(x[0], y[0]) - torch.matmul(x[1], y[1])
        z[1] = torch.matmul(x[0], y[1]) + torch.matmul(x[1], y[0])

    return z


def inner_prod(x, y):
    """A function that returns the inner product of two complex vectors,
    x and y >>> <x|y>.

    :param x: A complex vector.
    :type x: torch.doubleTensor
    :param y: A complex vector.
    :type y: torch.doubleTensor

    :raises ValueError: If x and y are not complex vectors with their first
                        dimensions being 2, then the function will not execute.

    :returns: The inner product, :math:`\\langle x\\vert y\\rangle`.
    :rtype: torch.doubleTensor
    """
    z = torch.zeros(2, dtype=torch.double, device=x.device)

    if len(list(x.size())) == 2 and len(list(y.size())) == 2:
        z[0] = torch.dot(x[0], y[0]) - torch.dot(-x[1], y[1])
        z[1] = torch.dot(x[0], y[1]) + torch.dot(-x[1], y[0])

    if len(list(x.size())) == 1 and len(list(y.size())) == 1:
        z[0] = x[0]*y[0] - (-x[1]*y[1])
        z[1] = x[0]*y[1] + (-x[1]*y[0])

    return z


def outer_prod(x, y):
    """A function that returns the outer product of two complex vectors, x and y.

    :param x: A complex vector.
    :type x: torch.doubleTensor
    :param y: A complex vector.
    :type y: torch.doubleTensor

    :raises ValueError:	If x and y are not complex vectors with their first
                        dimensions being 2, then the function will not execute.

    :returns: The outer product between x and y,
        :math:`\\vert x \\rangle\\langle y\\vert`.
    :rtype: torch.doubleTensor
    """
    if len(list(x.size())) < 2 or len(list(y.size())) < 2:
        raise ValueError('An input is not of the right dimension.')

    z = torch.zeros(2, x.size()[1], y.size()[1],
                    dtype=torch.double, device=x.device)
    z[0] = torch.ger(x[0], y[0]) - torch.ger(x[1], -y[1])
    z[1] = torch.ger(x[0], -y[1]) + torch.ger(x[1], y[0])

    return z


def conjugate(x):
    """A function that takes the complex transpose of the argument.

    :param x: A complex vector or matrix.
    :type x: torch.tensor

    :returns: The conjugate of x.
    :rtype: torch.tensor
    """
    if len(list(x.size())) == 2:
        z = torch.zeros(2, x.size()[1], dtype=torch.double, device=x.device)
        z[0] = x[0]
        z[1] = -x[1]

    if len(list(x.size())) == 3:
        z = torch.zeros(2, x.size()[2], x.size()[1],
                        dtype=torch.double, device=x.device)
        z[0] = torch.transpose(x[0], 0, 1)
        z[1] = -torch.transpose(x[1], 0, 1)

    return z


def kronecker_prod(x, y):
    """A function that returns the tensor / kronecker product of 2 comlpex
    tensors, x and y.

    :param x: A complex matrix.
    :type x: torch.doubleTensor
    :param y: A complex matrix.
    :type y: torch.doubleTensor

    :raises ValueError: If x and y do not have 3 dimensions or their first
                        dimension is not 2, the function cannot execute.

    :returns: The tensorproduct of x and y, :math:`x \\otimes y`.
    :rtype: torch.doubleTensor
    """
    if len(list(x.size())) != 3 or len(list(y.size())) != 3:
        raise ValueError('An input is not of the right dimension.')

    z = torch.zeros(2, x.size()[1]*y.size()[1],
                    x.size()[2]*y.size()[2],
                    dtype=torch.double, device=x.device)

    row_count = 0

    for i in range(x.size()[1]):
        for k in range(y.size()[1]):
            column_count = 0
            for j in range(x.size()[2]):
                for l in range(y.size()[2]):

                    z[0][row_count][column_count] = x[0][i][j] * \
                        y[0][k][l] - x[1][i][j]*y[1][k][l]
                    z[1][row_count][column_count] = x[0][i][j] * \
                        y[1][k][l] + x[1][i][j]*y[0][k][l]

                    column_count += 1
            row_count += 1

    return z


def scalar_divide(x, y):
    """A function that computes the division of x by y.

    :param x: The numerator (a complex scalar, vector or matrix).
    :type x: torch.tensor
    :param y: The denominator (a complex scalar).
    :type y: torch.tensor

    :returns: x / y
    :rtype: torch.tensor
    """
    if len(list(x.size())) == 2 or len(list(x.size())) == 1:
        y_star = torch.zeros_like(y)
        y_star[0] = y[0]
        y_star[1] = -y[1]

        numerator = scalar_mult(y_star, x)
        denominator = scalar_mult(y, y_star)[0]

    if len(list(x.size())) == 3:
        y_star = torch.zeros_like(y)
        y_star[0] = y[0]
        y_star[1] = -y[1]

        numerator = scalar_mult(y_star, x)
        denominator = scalar_mult(y, y_star)[0]

    return numerator / denominator


def norm(x):
    """A function that returns the norm of the argument.

    :param x: A complex scalar.
    :type x: torch.doubleTensor

    :returns: :math:`|x|^2`.
    :rtype: torch.doubleTensor
    """
    return inner_prod(x, x)[0]



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils import parameters_to_vector

from qucumber import _warn_on_missing_gpu

__all__ = [
    "BinaryRBM"
]


class BinaryRBM(nn.Module):
    def __init__(self, num_visible, num_hidden, zero_weights=False,
                 gpu=True, num_chains=100):
        super(BinaryRBM, self).__init__()
        self.num_visible = int(num_visible)
        self.num_hidden = int(num_hidden)
        self.num_chains = int(num_chains)
        self.num_pars = ((self.num_visible * self.num_hidden)
                         + self.num_visible + self.num_hidden)

        _warn_on_missing_gpu(gpu)
        self.gpu = gpu and torch.cuda.is_available()

        # Maximum number of visible units for exact enumeration
        self.size_cut = 16

        self.device = torch.device('cuda') if self.gpu else torch.device('cpu')

        if zero_weights:
            self.weights = nn.Parameter((torch.zeros(self.num_hidden,
                                                     self.num_visible,
                                                     device=self.device,
                                                     dtype=torch.double)),
                                        requires_grad=True)
            self.visible_bias = nn.Parameter(torch.zeros(self.num_visible,
                                                         device=self.device,
                                                         dtype=torch.double),
                                             requires_grad=True)
            self.hidden_bias = nn.Parameter(torch.zeros(self.num_hidden,
                                                        device=self.device,
                                                        dtype=torch.double),
                                            requires_grad=True)
        else:
            self.initialize_parameters()

    def __repr__(self):
        return ("BinaryRBM(num_visible={}, num_hidden={}, gpu={})"
                .format(self.num_visible, self.num_hidden, self.gpu))

    def initialize_parameters(self):
        """Randomize the parameters of the RBM"""
        self.weights = nn.Parameter(
            (torch.randn(self.num_hidden, self.num_visible,
                         device=self.device, dtype=torch.double)
             / np.sqrt(self.num_visible)), requires_grad=True)

        self.visible_bias = nn.Parameter(
            (torch.randn(self.num_visible,
                         device=self.device, dtype=torch.double)
             / np.sqrt(self.num_visible)),
            requires_grad=True)
        self.hidden_bias = nn.Parameter(
            (torch.randn(self.num_hidden,
                         device=self.device, dtype=torch.double)
             / np.sqrt(self.num_hidden)),
            requires_grad=True)

    def effective_energy(self, v):
        r"""The effective energies of the given visible states.

        .. math::

            \mathcal{E}(\bm{v}) &= -\sum_{j}b_j v_j
                        - \sum_{i}\log
                            \left\lbrack 1 +
                                  \exp\left(c_{i} + \sum_{j} W_{ij} v_j\right)
                            \right\rbrack

        :param v: The visible states.
        :type v: torch.Tensor

        :returns: The effective energies of the given visible states.
        :rtype: torch.Tensor
        """
        if len(v.shape) < 2:
            v = v.view(1, -1)
        visible_bias_term = torch.mv(v, self.visible_bias)
        hidden_bias_term = F.softplus(
            F.linear(v, self.weights, self.hidden_bias)
        ).sum(1)

        return -(visible_bias_term + hidden_bias_term)

    def effective_energy_gradient(self, v):
        """The gradients of the effective energies for the given visible states.

        :param v: The visible states.
        :type v: torch.Tensor

        :returns: 1d vector containing the gradients for all parameters
                  (computed on the given visible states v).
        :rtype: torch.Tensor
        """
        prob = self.prob_h_given_v(v)

        if len(v.shape) < 2:
            W_grad = -torch.einsum("j,k->jk", (prob, v))
            vb_grad = -v
            hb_grad = -prob
        else:
            W_grad = -torch.einsum("ij,ik->jk", (prob, v))
            vb_grad = -torch.einsum("ij->j", (v,))
            hb_grad = -torch.einsum("ij->j", (prob,))

        return parameters_to_vector([W_grad, vb_grad, hb_grad])

    def prob_v_given_h(self, h):
        """Given a hidden unit configuration, compute the probability
        vector of the visible units being on.

        :param h: The hidden unit
        :type h: torch.Tensor

        :returns: The probability of visible units being active given the
                  hidden state.
        :rtype: torch.Tensor
        """
        p = torch.sigmoid(F.linear(h, self.weights.t(), self.visible_bias))
        return p

    def prob_h_given_v(self, v):
        """Given a visible unit configuration, compute the probability
        vector of the hidden units being on.

        :param h: The hidden unit.
        :type h: torch.Tensor

        :returns: The probability of hidden units being active given the
                  visible state.
        :rtype: torch.Tensor
        """
        p = torch.sigmoid(F.linear(v, self.weights, self.hidden_bias))
        return p

    def sample_v_given_h(self, h):
        """Sample/generate a visible state given a hidden state.

        :param h: The hidden state.
        :type h: torch.Tensor

        :returns: Tuple containing prob_v_given_h(h) and the sampled visible
                  state.
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        p = self.prob_v_given_h(h)
        v = p.bernoulli()
        return v

    def sample_h_given_v(self, v):
        """Sample/generate a hidden state given a visible state.

        :param h: The visible state.
        :type h: torch.Tensor

        :returns: Tuple containing prob_h_given_v(v) and the sampled hidden
                  state.
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        p = self.prob_h_given_v(v)
        h = p.bernoulli()
        return h

    def compute_partition_function(self, space):
        """The natural logarithm of the partition function of the RBM.

        :param space: A rank 2 tensor of the visible space.
        :type space: torch.Tensor

        :returns: The natural log of the partition function.
        :rtype: torch.Tensor
        """
        neg_free_energies = -self.effective_energy(space)
        logZ = neg_free_energies.logsumexp(0)
        Z = logZ.exp()
        return Z



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from operator import itemgetter

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker

from .callback import Callback


class LivePlotting(Callback):
    r"""Plots metrics.
    This Callback is called at the end of each epoch.

    :param period: Frequency with which the callback updates the plots
                   (in epochs).
    :type period: int
    :param metric_callback: An instance of
        :class:`MetricEvaluator<MetricEvaluator>` which computes the metric
        that we want to check for convergence.
    :type metric_callback: :class:`MetricEvaluator<MetricEvaluator>`
    :param metric_name: The name of the metric stored in `metric_callback`.
    :type metric_name: str
    :param error_name: The name of the error stored in `metric_callback`.
    :type error_name: str
    """
    def __init__(self, period, metric_evaluator, metric_name,
                 error_name=None, total_epochs=None, smooth=True):
        self.period = period
        self.metric_evaluator = metric_evaluator
        self.metric_name = metric_name
        self.error_name = error_name
        self.last_epoch = 0
        self.total_epochs = total_epochs
        self.smooth = smooth

    def on_train_start(self, rbm):
        self.fig, self.ax = plt.subplots()

        if self.total_epochs:
            self.ax.set_xlim(0, self.total_epochs)

        self.ax.grid()
        self.ax.xaxis.set_major_locator(
            ticker.MultipleLocator(min(self.period, 5.))
        )
        self.fig.show()
        self.fig.canvas.draw()

    def on_epoch_end(self, rbm, epoch):
        if epoch % self.period == 0:
            self.last_epoch = epoch

            epochs = np.array(list(
                map(itemgetter(0), self.metric_evaluator.metric_values)
            ))

            metric_values = np.array(list(
                map(itemgetter(self.metric_name),
                    map(itemgetter(1), self.metric_evaluator.metric_values))
            ))

            self.ax.clear()
            p = self.ax.plot(epochs, metric_values)

            if self.error_name is not None:
                std_error = np.array(list(
                    map(itemgetter(self.error_name),
                        map(itemgetter(1),
                            self.metric_evaluator.metric_values))
                ))

                lower = metric_values - std_error
                upper = metric_values + std_error

                self.ax.fill_between(epochs, lower, upper,
                                     color=p[0].get_color(),
                                     alpha=0.4)

            y_avg = np.max(np.abs(metric_values))
            y_log_avg = np.log10(y_avg) if y_avg != 0 else -1.
            y_tick_exp = int(np.sign(y_log_avg) * np.ceil(np.abs(y_log_avg)))
            y_tick_interval = (10 ** y_tick_exp) / 2.
            self.ax.yaxis.set_major_locator(
                ticker.MultipleLocator(y_tick_interval)
            )

            self.ax.set_xlabel("Epochs")
            self.ax.set_ylabel(self.metric_name)
            self.ax.grid()
            self.fig.canvas.draw()

    def on_train_end(self, rbm):
        self.on_epoch_end(rbm, self.last_epoch)



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import torch
import numpy as np

def create_dict(name=None, unitary=None):
    dictionary = {
        'X': (1./np.sqrt(2))*torch.tensor(
                [[[1., 1.], [1., -1.]], [[0., 0.], [0., 0.]]],
                dtype=torch.double),
        'Y': (1./np.sqrt(2))*torch.tensor(
                [[[1., 0.], [1., 0.]], [[0., -1.], [0., 1.]]],
                dtype=torch.double),
        'Z': torch.tensor([[[1., 0.], [0., 1.]], [[0., 0.], [0., 0.]]],
                          dtype=torch.double)
    }

    if (name is not None) and (unitary is not None):
        dictionary[name] = unitary

    return dictionary



__version__ = "0.2.0"


def _warn_on_missing_gpu(gpu):
    import warnings

    import torch

    if gpu and not torch.cuda.is_available():
        warnings.warn("Could not find GPU: will continue with CPU.",
                      ResourceWarning)

def set_random_seed(seed, cpu=True, gpu=False, quiet=False):
    import warnings

    import torch

    if gpu and torch.cuda.is_available():
        if not quiet:
            warnings.warn("GPU random seeds are not completely deterministic. "
                          "Proceed with caution.")
        torch.cuda.manual_seed(seed)

    if cpu:
        torch.manual_seed(seed)



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from .callback import Callback


class Logger(Callback):
    """Callback which logs output at regular intervals.

    This Callback is called at the end of each epoch.

    :param period: Logging frequency (in epochs).
    :type period: int
    :param logger_fn: The function used for logging. Must take 1 string as
                      an argument. Defaults to the standard `print` function.
    :type logger_fn: callable
    :param msg_gen: A callable which generates the string to be logged.
                    Must take 2 positional arguments: the RBM being trained and
                    the current epoch. It must also be able to take some
                    keyword arguments.
    :type msg_gen: callable
    :param \**kwargs: Keyword arguments which will be passed to `msg_gen`.
    """
    def __init__(self, period, logger_fn=print,
                 msg_gen=None, **msg_gen_kwargs):
        self.period = period
        self.logger_fn = logger_fn
        self.msg_gen = msg_gen if callable(msg_gen) else self._default_msg_gen
        self.msg_gen_kwargs = msg_gen_kwargs

    @staticmethod
    def _default_msg_gen(rbm, epoch, **kwargs):
        return "Epoch " + str(epoch) + ": " + str(kwargs)

    def on_epoch_end(self, rbm, epoch):
        if epoch % self.period == 0:
            self.logger_fn(self.msg_gen(rbm, epoch, **self.msg_gen_kwargs))



from .callback import *
from .callback_list import *
from .early_stopping import *
from .lambda_callback import *
from .liveplotting import *
from .logger import *
from .metric_evaluator import *
from .model_saver import *
from .variance_based_early_stopping import *



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import torch
import numpy as np

def vector_to_grads(vec, parameters):
    r"""Convert one vector to the parameters

    Arguments:
        vec (Tensor): a single vector represents the parameters of a model.
        parameters (Iterable[Tensor]): an iterator of Tensors that are the
            parameters of a model.
    """
    # Ensure vec of type Tensor
    if not isinstance(vec, torch.Tensor):
        raise TypeError('expected torch.Tensor, but got: {}'
                        .format(torch.typename(vec)))
    # Flag for the device where the parameter is located
    param_device = None

    # Pointer for slicing the vector for each parameter
    pointer = 0
    for param in parameters:
        # Ensure the parameters are located in the same device
        param_device = _check_param_device(param, param_device)

        # The length of the parameter
        num_param = torch.prod(torch.LongTensor(list(param.size())))
        # Slice the vector, reshape it, and replace the old data of the parameter
        param.grad = vec[pointer:pointer + num_param].view(param.size()).data

        # Increment the pointer
        pointer += num_param


def _check_param_device(param, old_param_device):
    r"""This helper function is to check if the parameters are located
    in the same device. Currently, the conversion between model parameters
    and single vector form is not supported for multiple allocations,
    e.g. parameters in different GPUs, or mixture of CPU/GPU.

    Arguments:
        param ([Tensor]): a Tensor of a parameter of a model
        old_param_device (int): the device where the first parameter of a
                                model is allocated.

    Returns:
        old_param_device (int): report device for the first time
    """

    # Meet the first parameter
    if old_param_device is None:
        old_param_device = param.get_device() if param.is_cuda else -1
    else:
        warn = False
        if param.is_cuda:  # Check if in same GPU
            warn = (param.get_device() != old_param_device)
        else:  # Check if in CPU
            warn = (old_param_device != -1)
        if warn:
            raise TypeError('Found two parameters on different devices, '
                            'this is currently not supported.')
    return old_param_device



import setuptools
import qucumber

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
      name='qucumber',
      version=qucumber.__version__,
      description='Neural network quantum state tomography.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
      ],
      install_requires=[
        'torch==0.4.1',
        'tqdm>=4.23',
        'numpy>=1.14',
        'matplotlib>=2.2'
      ],
      include_package_data=True,
      url='http://github.com/PIQuIL/QuCumber',
      author='PIQuIL',
      author_email='piquildbeets@gmail.com',
      license='Apache License 2.0',
      packages=setuptools.find_packages(),
      zip_safe=False
)



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import warnings
from itertools import chain

import numpy as np
from math import sqrt
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm, tqdm_notebook

import qucumber.utils.cplx as cplx
from qucumber.callbacks import CallbackList
from qucumber.binary_rbm import BinaryRBM

__all__ = [
    "PositiveWavefunction"
]

class PositiveWavefunction(object):
    def __init__(self, num_visible, num_hidden=None, gpu=True):
        super(PositiveWavefunction, self).__init__()
        self.num_visible = int(num_visible)
        self.num_hidden = (int(num_hidden)
                           if num_hidden is not None
                           else self.num_visible)
        self.rbm_am = BinaryRBM(self.num_visible, self.num_hidden,
                                            gpu=gpu)

         # Maximum size of the Hilbert space for full enumeration
        self.size_cut = 20

        self.space = None
        self.Z = 0.0
        self.num_pars = self.rbm_am.num_pars
        self.networks = ["rbm_am"]
        self.device = self.rbm_am.device
        self.visible_state = torch.zeros(1,self.num_visible,
                                         device=self.rbm_am.device,
                                         dtype=torch.double)
        self.hidden_state = torch.zeros(1,self.num_hidden,
                                         device=self.rbm_am.device,
                                         dtype=torch.double)

    def initialize_parameters(self):
        r"""Randomize the parameters :math:`\bm{\lambda}=\{\bm{W},\bm{b},\bm{c}\}` of
        the RBM parametrizing the wavefunction."""
        self.rbm_am.initialize_parameters()

    def set_visible_layer(self,v):
        r""" Set the visible state to a given vector/matrix

        :param v: State to initialize the wavefunction to
        :type v: torch.Tensor
        """
        self.visible_state = v
        if (self.visible_state.shape != v.shape):
            raise RuntimeError ('Error in set_visible_layer')

    def psi(self,v):
        r""" Compute the wavefunction of a given vector/matrix of visible states:

        .. math::

            \psi_{\bm{\lambda}}(\bm{\sigma}) = e^{-\mathcal{E}_{\bm{\lambda}}(\bm{\sigma})/2}

        :param v: visible states :math:`\bm{\sigma}`
        :type v: torch.tensor

        :returns: Complex object containing the value of the wavefunction for each visible state
        :rtype: torch.tensor
        """
        psi = torch.zeros(2, dtype=torch.double, device = self.device)
        psi[0] = (-self.rbm_am.effective_energy(v)).exp().sqrt()#self.amplitude(v)
        psi[1] = 0.0
        return psi

    def gradient(self,v):
        r"""Compute the gradient :math:`\nabla_{\bm{\lambda}}\mathcal{E}_{\bm{\lambda}}(\bm{\sigma})` of the effective visible energy for a batch
        of visible states v.


        :param v: visible states :math:`\bm{\sigma}`
        :type v: torch.tensor

        :returns dictionary with one key (rbm_am)
        :rtype  dictionary(dictionary(torch.tensor,torch.tensor,torch.tensor)
        """
        return self.rbm_am.effective_energy_gradient(v)

    def sample(self, k_cd):
        r"""Performs k steps of Block Gibbs sampling. One step consists of sampling
        the hidden state :math:`\bm{h}` from the conditional distribution
        :math:`p_{\bm{\lambda}}(\bm{h}\:|\:\bm{v})`, and sampling the visible state
        :math:`\bm{v}` from the conditional distribution :math:`p_{\bm{\lambda}}(\bm{v}\:|\:\bm{h})`.

        :param k_cd: Number of Block Gibbs steps.
        :type k_cd: int
        """
        for _ in range(k_cd):
            self.hidden_state = self.rbm_am.sample_h_given_v(self.visible_state)
            self.visible_state = self.rbm_am.sample_v_given_h(self.hidden_state)

    def save(self, location, metadata={}):
        """Saves the RBM parameters to the given location along with
        any given metadata.

        :param location: The location to save the RBM parameters + metadata
        :type location: str or file
        :param metadata: Any extra metadata to store alongside the RBM
                         parameters
        :type metadata: dict
        """
        # add extra metadata to dictionary before saving it to disk
        data = {"rbm_am":self.rbm_am.state_dict(), **metadata}
        torch.save(data, location)

    def load(self, location):
        """Loads the RBM parameters from the given location ignoring any
        metadata stored in the file. Overwrites the RBM's parameters.

        .. note::
            The RBM object on which this function is called must
            have the same shape as the one who's parameters are being
            loaded.

        :param location: The location to load the RBM parameters from
        :type location: str or file
        """
        try:
            state_dict = torch.load(location)
        except AssertionError as e:
            state_dict = torch.load(location, lambda storage, loc: 'cpu')

        self.rbm_am.load_state_dict(state_dict, strict=False)

    def generate_Hilbert_space(self,size):
        r"""Generates Hilbert space of dimension :math:`2^{\text{size}}`.

        :returns: A tensor with all the basis states of the Hilbert space.
        :rtype: torch.Tensor
        """
        if (size > self.size_cut):
            raise ValueError('Size of the Hilbert space too large!')
        else:
            space = torch.zeros((1 << size, size),
                                device=self.device, dtype=torch.double)
            for i in range(1 << size):
                d = i
                for j in range(size):
                    d, r = divmod(d, 2)
                    space[i, size - j - 1] = int(r)
            return space

    def compute_normalization(self):
        r"""Compute the normalization constant of the wavefunction.

        .. math::

            Z_{\bm{\lambda}}=\sqrt{\sum_{\bm{\sigma}}|\psi_{\bm{\lambda}}|^2}=
            \sqrt{\sum_{\bm{\sigma}} p_{\bm{\lambda}}(\bm{\sigma})}

        :param space: A rank 2 tensor of the entire visible space.
        :type space: torch.Tensor

        """
        if (self.space is None):
            raise ValueError('Missing Hilbert space')
        else:
            self.Z = self.rbm_am.compute_partition_function(self.space)

    @staticmethod
    def autoload(location, gpu=False):
        """Initializes an RBM from the parameters in the given location,
        ignoring any metadata stored in the file.

        :param location: The location to load the RBM parameters from
        :type location: str or file

        :returns: A new RBM initialized from the given parameters
        :rtype: BinomialRBM
        """
        state_dict = torch.load(location)

        rbm = BinaryRBM(num_visible=len(state_dict["rbm_am"]['visible_bias']),
                        num_hidden=len(state_dict["rbm_am"]['hidden_bias']),
                        gpu=gpu,
                        seed=None)

        rbm.load_state_dict(state_dict, strict=False)

        return rbm



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import torch
import numpy as np
import qucumber.utils.cplx as cplx
import qucumber.utils.unitaries as unitaries

def fidelity(nn_state,target_psi,bases=None):
    nn_state.compute_normalization() 
    F = torch.tensor([0., 0.], dtype=torch.double, device = nn_state.device) 
    target_psi = target_psi.to(nn_state.device)
    for i in range(len(nn_state.space)):
        psi = nn_state.psi(nn_state.space[i])/(nn_state.Z).sqrt()
        F[0] += target_psi[0,i]*psi[0]+target_psi[1,i]*psi[1]
        F[1] += target_psi[0,i]*psi[1]-target_psi[1,i]*psi[0]
    return  cplx.norm(F)

def rotate_psi(nn_state,basis,unitaries,psi=None):
    N=nn_state.num_visible
    v = torch.zeros(N, dtype=torch.double, device=nn_state.device)
    psi_r = torch.zeros(2,1<<N,dtype=torch.double,device = nn_state.device)
    for x in range(1<<N):
        Upsi = torch.zeros(2, dtype=torch.double, device=nn_state.device)
        num_nontrivial_U = 0
        nontrivial_sites = []
        for j in range(N):
            if (basis[j] != 'Z'):
                num_nontrivial_U += 1
                nontrivial_sites.append(j)
        sub_state = nn_state.generate_Hilbert_space(num_nontrivial_U)
         
        for xp in range(1<<num_nontrivial_U):
            cnt = 0
            for j in range(N):
                if (basis[j] != 'Z'):
                    v[j]=sub_state[xp][cnt] 
                    cnt += 1
                else:
                    v[j]=nn_state.space[x,j]
            U = torch.tensor([1., 0.], dtype=torch.double, device=nn_state.device)
            for ii in range(num_nontrivial_U):
                tmp = unitaries[basis[nontrivial_sites[ii]]][:,int(nn_state.space[x][nontrivial_sites[ii]]),int(v[nontrivial_sites[ii]])].to(nn_state.device)
                U = cplx.scalar_mult(U,tmp)
            if psi is None:
                Upsi += cplx.scalar_mult(U,nn_state.psi(v))
            else:
                index = 0
                for k in range(len(v)):
                    index = (index << 1) | int(v[k].item())
                Upsi += cplx.scalar_mult(U,psi[:,index])
        psi_r[:,x] = Upsi
    return psi_r


def KL(nn_state,target_psi,bases=None):
    psi_r = torch.zeros(2,1<<nn_state.num_visible,dtype=torch.double, device = nn_state.device)
    KL = 0.0
    unitary_dict = unitaries.create_dict()
    target_psi = target_psi.to(nn_state.device)
    space = nn_state.generate_Hilbert_space(nn_state.num_visible)
    nn_state.compute_normalization()
    if bases is None:
        num_bases = 1
        for i in range(len(space)):
            KL += cplx.norm(target_psi[:,i])*cplx.norm(target_psi[:,i]).log()
            KL -= cplx.norm(target_psi[:,i])*cplx.norm(nn_state.psi(space[i])).log()
            KL += cplx.norm(target_psi[:,i])*nn_state.Z.log()
            
    else:
        num_bases = len(bases)
        for b in range(1,len(bases)):
            psi_r = rotate_psi(nn_state,bases[b],unitary_dict)
            target_psi_r = rotate_psi(nn_state,bases[b],unitary_dict,target_psi)
            for ii in range(len(space)):
                if(cplx.norm(target_psi_r[:,ii])>0.0):
                    KL += cplx.norm(target_psi_r[:,ii])*cplx.norm(target_psi_r[:,ii]).log()
                KL -= cplx.norm(target_psi_r[:,ii])*cplx.norm(psi_r[:,ii]).log().item()
                KL += cplx.norm(target_psi_r[:,ii])*nn_state.Z.log()
    return  KL/float(num_bases)    

def load_target_data(psi_data,bases_data):
    D = int(len(psi_data)/float(len(bases_data)))
    target_psi_dict = {}
    bases = []
    for i in range(len(bases_data)):
        tmp = ""
        for j in range(len(bases_data[i])):
            if bases_data[i][j] is not " ":
                tmp += bases_data[i][j]
        bases.append(tmp)
    for b in range(len(bases)):
        psi      = torch.zeros(2,D, dtype=torch.double)
        psi_real = torch.tensor(psi_data[b*D:D*(b+1),0], dtype=torch.double)
        psi_imag = torch.tensor(psi_data[b*D:D*(b+1),1], dtype=torch.double)
        psi[0]   = psi_real
        psi[1]   = psi_imag
        
        target_psi_dict[bases[b]] = psi
    target_psi = target_psi_dict[bases[0]]
    
    return bases,target_psi
   



# Copyright 2018 PIQuIL - All Rights Reserved

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#import warnings
from itertools import chain

import numpy as np
from math import sqrt
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from torch.nn.utils import parameters_to_vector
from tqdm import tqdm, tqdm_notebook
import time

import qucumber.utils.cplx as cplx
from qucumber.utils.gradients_utils import vector_to_grads,_check_param_device
from qucumber.callbacks import CallbackList
from qucumber.positive_wavefunction import PositiveWavefunction
from qucumber.complex_wavefunction import ComplexWavefunction
import math as m

__all__ = [
    "QuantumReconstruction"
]

class QuantumReconstruction(object):
    def __init__(self, nn_state):
        super(QuantumReconstruction, self).__init__()
        self.nn_state = nn_state 
        self.num_visible = nn_state.num_visible
        self.stop_training = False


    def compute_batch_gradients(self, k_cd, samples_batch,bases_batch=None):
        """This function will compute the gradients of a batch of the training
        data (samples_batch). If measurements are taken in bases other than the
        reference basis, a list of bases (bases_batch) must also be provided.
        
        :param k_cd: Number of contrastive divergence steps in training.
        :type k_cd: int
        :param samples_batch: Batch of the input samples.
        :type samples_batch: torch.Tensor
        :param bases_batch: Batch of the input bases 
        :type bases_batch: np.array

        :returns: List containing the gradients of the parameters.
        :rtype: list
        """
        # Negative phase: learning signal driven by the amplitude RBM of the NN state
        self.nn_state.sample(k_cd)
        grad_model = self.nn_state.rbm_am.effective_energy_gradient(self.nn_state.visible_state)


        # If measurements are taken in the reference bases only
        if bases_batch is None:
            grad = [0.0]
            # Positive phase: learning signal driven by the data (and bases)
            grad_data = self.nn_state.gradient(samples_batch)
            # Gradient = Positive Phase - Negative Phase        
            grad[0] = grad_data/float(samples_batch.shape[0]) - grad_model/float(self.nn_state.visible_state.shape[0])
        else:
            grad=[0.0,0.0]
            # Initialize
            grad_data = [torch.zeros(self.nn_state.rbm_am.num_pars,dtype=torch.double, device=self.nn_state.device),torch.zeros(self.nn_state.rbm_ph.num_pars,dtype=torch.double, device=self.nn_state.device)]
            # Loop over each sample in the batch
            for i in range(samples_batch.shape[0]):
                # Positive phase: learning signal driven by the data (and bases)
                data_gradient = self.nn_state.gradient(bases_batch[i],samples_batch[i])
                grad_data[0] += data_gradient[0] #Accumulate amplitude RBM gradient
                grad_data[1] += data_gradient[1] #Accumulate phase RBM gradient
            
            # Gradient = Positive Phase - Negative Phase
            grad[0] = grad_data[0]/float(samples_batch.shape[0]) - grad_model/float(self.nn_state.visible_state.shape[0])
            grad[1] = grad_data[1]/float(samples_batch.shape[0])    #No negative signal for the phase parameters
        
        return grad
        
    def fit(self,input_samples,epochs,pos_batch_size, neg_batch_size,k_cd,lr,
            observer=None,
            input_bases = None,z_samples = None,
            progbar=False,callbacks=[]):
        """Execute the training of the RBM.

        :param input_samples: The training samples
        :type input_samples: np.array
        :param epochs: 
        :type epochs: int
        :param pos_batch_size: The size of batches for the positive phase
                               taken from the data.
        :type pos_batch_size: int
        :param neg_batch_size: The size of batches for the negative phase
                               taken from the data
        :type neg_batch_size: int
        :param k: The number of contrastive divergence steps
        :type k: int
        :param lr: Learning rate
        :type lr: float
        :param progbar: Whether or not to display a progress bar. If "notebook"
                        is passed, will use a Jupyter notebook compatible
                        progress bar.
        :type progbar: bool or str
        :param callbacks: Callbacks to run while training.
        :type callbacks: list(qucumber.callbacks.Callback)
        """
        disable_progbar = (progbar is False)
        progress_bar = tqdm_notebook if progbar == "notebook" else tqdm
        callbacks = CallbackList(callbacks)

        train_samples = torch.tensor(input_samples, device=self.nn_state.device,
                            dtype=torch.double)
        
        if (len(self.nn_state.networks) >1):
            optimizer = torch.optim.SGD(list(self.nn_state.rbm_am.parameters())+list(self.nn_state.rbm_ph.parameters()),lr=lr)

        else:
            optimizer = torch.optim.SGD(self.nn_state.rbm_am.parameters(),lr=lr)
            batch_bases = None
        callbacks.on_train_start(self.nn_state)
        t0 = time.time()
        batch_num = m.ceil(train_samples.shape[0] / pos_batch_size)
        for ep in progress_bar(range(1,epochs+1), desc="Epochs ",
                               disable=disable_progbar):

            random_permutation      = torch.randperm(train_samples.shape[0])
            shuffled_samples        = train_samples[random_permutation]
            # List of all the batches for positive phase.
            pos_batches = [shuffled_samples[batch_start:(batch_start + pos_batch_size)]
                           for batch_start in range(0, len(train_samples), pos_batch_size)]

            if input_bases is not None:
                shuffled_bases = input_bases[random_permutation]
                pos_batches_bases = [shuffled_bases[batch_start:(batch_start + pos_batch_size)]
                           for batch_start in range(0, len(train_samples), pos_batch_size)]
             
            callbacks.on_epoch_start(self.nn_state, ep)

            if self.stop_training:  # check for stop_training signal
                break
            
            for b in range(batch_num):
                callbacks.on_batch_start(self.nn_state, ep, b)

                if input_bases is None:
                    random_permutation      = torch.randperm(train_samples.shape[0])
                    neg_batch = train_samples[random_permutation][0:neg_batch_size]
                
                else:
                    z_samples = z_samples.to(self.nn_state.device)
                    random_permutation      = torch.randperm(z_samples.shape[0])
                    neg_batch = z_samples[random_permutation][0:neg_batch_size]
                    batch_bases = pos_batches_bases[b]

                self.nn_state.set_visible_layer(neg_batch)
                all_grads = self.compute_batch_gradients(k_cd, pos_batches[b],batch_bases)
                optimizer.zero_grad()  # clear any cached gradients
                
                for p,net in enumerate(self.nn_state.networks):
                    rbm = getattr(self.nn_state,net)
                    vector_to_grads(all_grads[p],rbm.parameters())
                optimizer.step()  # tell the optimizer to apply the gradients

                callbacks.on_batch_end(self.nn_state, ep, b)

            callbacks.on_epoch_end(self.nn_state, ep)

        t1 = time.time()
        print("\nElapsed time = %.2f" %(t1-t0)) 
